import abc
from typing import Optional

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    import torch

# Global context
__global_attention_context = None

def attention_context():
    return __global_attention_context

def clear_attention_context():
    global __global_attention_context
    __global_attention_context = None

def set_attention_context(context: "AttentionInterceptionContext"):
    global __global_attention_context
    __global_attention_context = context


class AttentionInterceptionContext(abc.ABC):
    """
    A basic context to capture attention scores in the models.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __enter__(self):
        set_attention_context(self)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        clear_attention_context()
        self.reset()

    def reset(self):
        pass

    def intercept_attention_weights(
        self,
        module: "AttentionInterceptor",
        attention_weights: "torch.Tensor"
    ) -> "torch.Tensor":
        return attention_weights


class AttentionInterceptor:
    """
    A mixin for scaled dot product attention classes to provide attention scores to
    the current attention context. Inherit from this mixin, and then
    use self.log_attention_weights(attention_weights) to supply the weights to the context.
    Currently, minibatched attention is not supported.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def attention_context(self) -> Optional["AttentionInterceptionContext"]:
        return attention_context()

    def intercept_attention_weights(self, attention_weights) -> "torch.Tensor":
        context = attention_context()
        if context is None:
            return attention_weights
        return context.intercept_attention_weights(self, attention_weights)
