from typing import Any, Callable, Optional, overload, Union
from ust.api.types import Minibatches
from ust.checkpointing import MinibatchCheckpoint, MinibatchContext, minibatch_context

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    import torch

class MinibatchCheckpointableMixin:
    """
    A mixin that provides minibatch checkpointing. If no minibatch context is available,
    these methods are no-ops.
    """
    @property
    def minibatch_context(self) -> Optional[MinibatchContext]:
        """
        Get the current minibatch context.
        """
        return minibatch_context()

    def minibatch_checkpoint(self, key: str = "") -> Optional[MinibatchCheckpoint]:
        """
        Get the current minibatch checkpoint.
        """
        minibatch_context = self.minibatch_context
        if minibatch_context is None:
            return None
        return minibatch_context.checkpoint(f"{id(self)}.{key}")

    @overload
    def minibatch_checkpoint_pipe(
        self,
        x: "torch.Tensor",
        key: str = "",
    ) -> "torch.Tensor":
        ...

    @overload
    def minibatch_checkpoint_pipe(
        self,
        x: Minibatches,
        key: str = "",
        extract_reference_tensor: Optional[Callable[[Any], "torch.Tensor"]] = None
    ) -> Minibatches:
        ...

    def minibatch_checkpoint_pipe(
        self,
        x: Union["torch.Tensor", Minibatches],
        key: str = "",
        extract_reference_tensor: Optional[Callable[[Any], "torch.Tensor"]] = None
    ):
        """
        Mark the region in the network as a checkpoint and feed the minibatches through this
        checkpoint. This will ensure that gradients are stopped according to the current
        minibatch context.
        """
        minibatch_context = self.minibatch_context
        if minibatch_context is None:
            return x
        return minibatch_context.checkpoint_pipe(
            x,
            f"{id(self)}.{key}",
            extract_reference_tensor=extract_reference_tensor
        )

    def set_minibatch_checkpoint_reference_tensor(
        self,
        x: "torch.Tensor",
        key: str = ""
    ):
        """
        Set the reference tensor for the minibatch checkpoint. The reference tensor is used
        to find parameters that need to be scaled for unbiased gradient estimation.
        """
        minibatch_checkpoint = self.minibatch_checkpoint(key=key)
        if minibatch_checkpoint is None:
            return
        minibatch_checkpoint.observed_reference = x
