"""
An implementation of Self Attention Attribution (https://arxiv.org/abs/2004.11207)
"""
import torch
from tqdm import tqdm
from typing import Any, Callable, Dict, Iterable, List, Optional

from .api.attention import AttentionInterceptionContext, AttentionInterceptor

class AttentionAttributionContext(AttentionInterceptionContext):
    def __init__(self, modules: Iterable["AttentionInterceptor"]):
        super().__init__()
        self.modules: List["AttentionInterceptor"] = list(modules)
        self.attention_weights: Dict["AttentionInterceptor", Optional[torch.Tensor]] = {}
        self.attention_head_mask = 1.0

    def reset(self):
        self.attention_weights.clear()

    def intercept_attention_weights(
        self,
        module: "AttentionInterceptor",
        attention_weights: torch.Tensor
    ) -> torch.Tensor:
        if module not in self.attention_weights:
            # Skip logs from other modules
            return attention_weights
        if self.attention_weights[module] is not None:
            raise ValueError(f"Found duplicate module {module} (id: {module})")
        self.attention_weights[module] = self.attention_head_mask*attention_weights
        return self.attention_weights[module] # type: ignore

    def compute(
        self,
        objective: Callable[[Any], torch.Tensor],
        inputs,
        integration_steps: int = 20,
        show_progress: bool = True
    ):
        """
        Returns:
            A list of attribution scores for each module. Each attribution score is a tensor of shape
            (batch_size, num_heads, query_len, key_len).
        """
        # Setup first forward pass, capturing all attention weights.
        self.attention_weights = {module: None for module in self.modules}
        self.attention_head_mask = 1.0
        loss = objective(inputs)
        attention_weights = list(self.attention_weights.values()) # type: ignore
        assert all([w is not None for w in attention_weights]), "Failed to capture all attention weights"
        attribution_results = [g.detach() for g in torch.autograd.grad(loss, attention_weights)] # type: ignore
        self.attention_weights.clear()

        # Integration module-by-module
        for module, A, attribution in zip(tqdm(self.modules, disable=(not show_progress)), attention_weights, attribution_results):
            # Integrate
            for k in torch.linspace(0.0, 1.0, integration_steps)[:-1]:
                # Add module as a target for intercepting attention weights
                self.attention_weights[module] = None
                self.attention_head_mask = k
                y = objective(inputs)
                attention_weights: torch.Tensor = self.attention_weights[module] # type: ignore
                assert attention_weights is not None, f"Failed to capture attention weights for module {module}"
                attribution += torch.autograd.grad(y, attention_weights)[0].detach()
            attribution[:] = A * attribution / integration_steps

            # Remove module_id as a target for intercepting attention weights
            del self.attention_weights[module]

        # [num_modules, batch_size, num_heads, query_len, key_len]
        return attribution_results


# class AttentionAttributionProvider(abc.ABC, nn.Module):
#     def get_attention_modules(self) -> List["AttentionInterceptor"]:
#         self.attention_modules = []
#         for module in self.modules():
#             if isinstance(module, AttentionInterceptor):
#                 self.attention_modules.append(module)
#         return self.attention_modules

#     @abc.abstractmethod
#     def aggregate_attention_scores(self, attention_scores: List[torch.Tensor]):
#         raise NotImplementedError


# class SelfAttentionAttributionProvider(AttentionAttributionProvider):
#     def aggregate_attention_scores(self, attention_scores: List[torch.Tensor]):
#         assert len(attention_scores) == 1
#         return attention_scores[0]


# class CrossAttentionAttributionProvider(AttentionAttributionProvider):
#     def get_attention_modules(self) -> List["AttentionInterceptor"]:
#         assert len(self.attention_modules) == 1
#         return self.attention_modules


# class InducedAttentionAttributionProvider(AttentionAttributionProvider):
#     def get_attention_modules(self) -> List["AttentionInterceptor"]:
#         raise NotImplementedError
