from dataclasses import dataclass
import torch
from typing import Any, Callable, Iterable, Literal, List, Optional

@dataclass
class MinibatchCheckpoint:

    # The number of minibatches that have been observed
    # at this checkpoint for the current context
    num_observed: int

    # A reference tensor from the minibatches.
    # This is used to find parameters that need to be scaled.
    observed_reference: Any

    params: Optional[List[torch.nn.Parameter]] = None

    def reset(self, reset_params: bool):
        self.num_observed = 0
        self.observed_reference = None
        if reset_params:
            self.params = None

    def pipe_immediate(
        self,
        tensor: torch.Tensor,
        gradient_minibatch_limit: Optional[int]
    ):
        n = gradient_minibatch_limit or float("inf")

        # Disable gradients if we are past the gradient limit
        if self.num_observed >= n:
            tensor = tensor.detach()

        # Update the info
        self.num_observed += 1
        if self.observed_reference is None:
            self.observed_reference = tensor

        return tensor

    def pipe_minibatch(
        self,
        minibatches: Iterable[Any],
        gradient_minibatch_limit: Optional[int],
        extract_reference_tensor: Optional[Callable[[Any], torch.Tensor]] = None
    ):
        n = gradient_minibatch_limit or float("inf")

        # This is the original gradient state at the start of processing
        original_gradient_state = torch.is_grad_enabled()

        # The is the last known gradient state
        prev_gradient_state = original_gradient_state

        # Disable gradients if we are past the gradient limit
        if prev_gradient_state and self.num_observed >= n:
            torch.set_grad_enabled(False)

        try:
            # Pull one minibatch at a time
            for i, mb in enumerate(minibatches, start=self.num_observed):
                if not isinstance(mb, torch.Tensor) and i >= n:
                    # Force detach gradients
                    for k, v in mb.items():
                        if isinstance(v, torch.Tensor):
                            mb[k] = v.detach()
                elif isinstance(mb, torch.Tensor) and i >= n:
                    mb = mb.detach()

                # Toggle gradients to the previous state before returning
                if torch.is_grad_enabled() != prev_gradient_state:
                    torch.set_grad_enabled(prev_gradient_state)

                # Update the info
                self.num_observed += 1
                if self.observed_reference is None:
                    if extract_reference_tensor is not None:
                        self.observed_reference = extract_reference_tensor(mb)
                    else:
                        self.observed_reference = mb
                # Yield the item
                yield mb

                # Get the current gradient state as we return back.
                # We'll use this to go back to this state before yielding.
                prev_gradient_state = torch.is_grad_enabled()

                # Disable gradients if we reach/pass the gradient limit
                if torch.is_grad_enabled() and i + 1 >= n:
                    torch.set_grad_enabled(False)

                # If we are allowed to turn gradients on, only do so if we aren't in the original state
                elif i < n and torch.is_grad_enabled() != original_gradient_state:
                    torch.set_grad_enabled(original_gradient_state)
        finally:
            torch.set_grad_enabled(prev_gradient_state)

    def __call__(
        self,
        minibatches: Iterable[Any],
        gradient_minibatch_limit: Optional[int],
        extract_reference_tensor: Optional[Callable[[Any], torch.Tensor]] = None
    ):
        if isinstance(minibatches, torch.Tensor):
            return self.pipe_immediate(minibatches, gradient_minibatch_limit)
        else:
            return self.pipe_minibatch(minibatches, gradient_minibatch_limit, extract_reference_tensor)


class MinibatchContext:
    _context = None

    def __init__(
        self,
        gradient_minibatch_limit: Optional[int] = None,
        gradient_estimation_method: Literal["biased", "unbiased"] = "biased",
        persist_graph: bool = False
    ):
        assert gradient_estimation_method in ("biased", "unbiased")
        self.gradient_minibatch_limit = gradient_minibatch_limit
        self.gradient_estimation_method = gradient_estimation_method
        self.persist_graph = persist_graph
        self._mapped_params = False
        self.checkpoints = {}

    def checkpoint_pipe(self, minibatches: Iterable[Any], key: str, extract_reference_tensor: Optional[Callable[[Any], torch.Tensor]] = None):
        """
        Pipe the minibatches through the checkpoint.
        """
        checkpoint = self.checkpoint(key)
        return checkpoint(
            minibatches,
            self.gradient_minibatch_limit,
            extract_reference_tensor=extract_reference_tensor
        )

    def checkpoint(self, key: Any) -> MinibatchCheckpoint:
        """
        Get a checkpoint
        """
        if key not in self.checkpoints:
            self.checkpoints[key] = MinibatchCheckpoint(0, None)
        return self.checkpoints[key]

    def print_model_with_scaling(self, model):
        """
        Print the model architecture like torch does, and for each module print
        its attribute name in the parent, its direct parameters side-by-side
        with their checkpoint scalar, and never print the same parameter twice.
        """
        # Build a lookup from parameter id to its checkpoint
        checkpoint_params = {
            id(p): checkpoint
            for checkpoint in self.checkpoints.values()
            for p in (checkpoint.params or [])
        }

        seen = set()

        def _traverse(module: torch.nn.Module, prefix: str = "", indent: int = 0):
            # Module header: include prefix if present
            cls_name = module.__class__.__name__
            extra = module.extra_repr()
            header = f"{cls_name}({extra})" if extra else cls_name
            line = f"{prefix + ': ' if prefix else ''}{header}"

            # Collect this module’s own parameters
            params = list(module.named_parameters(recurse=False))
            param_strs = []
            for pname, p in params:
                if id(p) in seen:
                    continue
                cp = checkpoint_params.get(id(p))
                # if a parameter isn't in any checkpoint, cp will be None
                observed = cp.num_observed if cp is not None else 0
                observed_with_gradients = min(observed, self.gradient_minibatch_limit or 0)
                param_strs.append(f"({pname}: {observed}/{observed_with_gradients})")
                seen.add(id(p))

            # Print with indentation
            indent_str = " " * indent
            if param_strs:
                print(f"{indent_str}{line}  " + "  ".join(param_strs))
            else:
                print(f"{indent_str}{line}")

            # Recurse into named children so we can capture their attribute names
            for child_name, child in module.named_children():
                _traverse(child, prefix=child_name, indent=indent + 2)

        # start with no prefix (the root module)
        _traverse(model)

    def correct_gradients(self):
        """
        Scale the gradients in the model. Run this after gradients have been computed on the tensors.
        """
        if self.gradient_estimation_method == "biased" or len(self.checkpoints) == 0:
            return

        n = self.gradient_minibatch_limit or 1

        # Create the parameter map if it doesn't exist
        if not self._mapped_params:
            self.map_checkpoints_to_parameters()

        for i, (key, ck) in enumerate(self.checkpoints.items()):
            if ck.params is None:
                raise RuntimeError(f"Checkpoint {key!r} has no parameters.")
            for p in ck.params:
                assert ck.num_observed > 0
                p.grad.data.mul_(ck.num_observed/min(ck.num_observed, n))

    def map_checkpoints_to_parameters(self):
        assigned_params = set()

        for key, ck in self.checkpoints.items():
            ref = ck.observed_reference
            if ref is None:
                raise RuntimeError(f"Checkpoint {key!r} has no reference tensor.")

            fn = ref.grad_fn
            if fn is None:
                raise RuntimeError(
                    f"Checkpoint {key!r} has no grad_fn; "
                    "ensure you passed a non-leaf Tensor."
                )

            seen_fns = set()
            reachable = set()

            def walk(f):
                if f in seen_fns:
                    return
                seen_fns.add(f)
                for next_fn, _ in f.next_functions:
                    if next_fn is None:
                        continue
                    # AccumulateGrad nodes carry .variable
                    if hasattr(next_fn, "variable"):
                        var = next_fn.variable
                        if isinstance(var, torch.nn.Parameter) or (isinstance(var, torch.Tensor) and var.requires_grad):
                            reachable.add(var)
                    else:
                        walk(next_fn)

            walk(fn)

            # Filter out any params already assigned upstream
            new_params = reachable - assigned_params
            ck.params = list(new_params)

            # Mark these as claimed
            assigned_params.update(new_params)

        self._mapped_params = True

    def reset(self):
        """
        Reset the context.
        """
        for checkpoint in self.checkpoints.values():
            checkpoint.reset(reset_params=not self.persist_graph)
        if not self.persist_graph:
            self._mapped_params = False

    def __enter__(self):
        self.__class__._context = self
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.__class__._context = None
        self.reset()


def minibatch_context():
    """
    Get the current minibatch context.
    """
    return MinibatchContext._context
