import math
import torch
import torch.nn.functional as F
from typing import Any, Optional
import warnings

from ..api.attention import AttentionInterceptor
from ..api.modules import ScaledDotProductSetAttention

class ScaledDotProductSoftmaxSetAttention(AttentionInterceptor, ScaledDotProductSetAttention):
    def compute_immediate_attention(
        self,
        query: torch.Tensor,
        key: torch.Tensor,
        value: torch.Tensor,
        key_multiplicities: Optional[torch.Tensor]
    ):
        scale = self.scale or 1.0/math.sqrt(query.shape[-1])
        attention = query @ key.transpose(-2, -1) * scale
        if key_multiplicities is not None:
            attention = attention + torch.log(key_multiplicities)
        attention = F.softmax(attention, dim=-1)
        attention = torch.dropout(attention, self.dropout_p, self.training)
        attention = self.intercept_attention_weights(attention)
        return attention @ value

    def compute_minibatch_attention(
        self,
        query: torch.Tensor,
        key: torch.Tensor,
        value: torch.Tensor,
        key_multiplicities: Optional[torch.Tensor]
    ):
        scale = self.scale or 1.0/math.sqrt(query.shape[-1])
        attention = query @ key.transpose(-2, -1) * scale
        if key_multiplicities is not None:
            attention = attention + torch.log(key_multiplicities.float()).unsqueeze(-2)
        max_value = attention.amax(dim=-1, keepdim=True)
        attention = torch.exp(attention - max_value)
        #Zero NaNs that might occur with empty sets (i.e. no multiplicities)
        attention = attention.masked_fill(torch.isnan(attention), 0.0)
        return {
            "attention": torch.dropout(attention, self.dropout_p, self.training) @ value,
            "normalization_term": attention.sum(dim=-1, keepdim=True),
            "max_value": max_value
        }

    def aggregate_minibatch_attention(
        self,
        src_state: Any,
        sink_state: Any
    ):
        # Get new state info
        max_value = src_state["max_value"]

        # If we have a previous state, we need to rescale before merging
        if sink_state["max_value"] is not None:
            # Scale sink state down to fit in context of src state
            # i.e. src state is bigger
            with torch.no_grad():
                # 1.0 when sink_state is bigger, <1.0 when sink state is smaller
                delta = torch.exp(torch.clamp(sink_state["max_value"] - max_value, max=0.0))
            sink_state["attention"] = sink_state["attention"]*delta
            sink_state["normalization_term"] = sink_state["normalization_term"]*delta

            # Scale src state down to fit in context of the sink state
            # i.e. sink state is bigger
            with torch.no_grad():
                delta = torch.exp(torch.clamp(max_value - sink_state["max_value"], max=0.0))
            src_state["attention"] = src_state["attention"]*delta
            src_state["normalization_term"] = src_state["normalization_term"]*delta

            max_value = torch.maximum(max_value, sink_state["max_value"])

        # Update state with new values
        sink_state["attention"] = sink_state["attention"] + src_state["attention"]
        sink_state["normalization_term"] = sink_state["normalization_term"] + src_state["normalization_term"]
        sink_state["max_value"] = max_value

    def initial_state(self):
        return {
            "attention": 0.0,
            "normalization_term": 0.0,
            "max_value": None
        }

    def get(self, state: Any):
        output = state["attention"] / state["normalization_term"]
        # Zero-out any NaNs that might occur if normalization_term is zero.
        # This can happen for an empty set (i.e. 0 for all multiplicities)
        return output.masked_fill(output.isnan(), 0.0)


class ScaledDotProductSoftmaxFlashSetAttention(ScaledDotProductSoftmaxSetAttention):
    def compute_immediate_attention(
        self,
        query: torch.Tensor,
        key: torch.Tensor,
        value: torch.Tensor,
        key_multiplicities: Optional[torch.Tensor]
    ):
        # If there is an attention context requesting scores, we need to do manual computation
        if self.attention_context is not None:
            return super().compute_immediate_attention(query, key, value, key_multiplicities)
        # Use native scaled dot product attention with attention bias
        attn_bias = None
        if key_multiplicities is not None:
            attn_bias = torch.log(key_multiplicities)
        output = F.scaled_dot_product_attention(
            query,
            key,
            value,
            attn_mask = attn_bias,
            dropout_p = self.dropout_p,
            is_causal = False,
            scale = self.scale,
            enable_gqa = self.enable_gqa
        )
        # Zero NaNs that might occur with empty sets or numerical issues
        return output.masked_fill(output.isnan(), 0.0)

    def compute_minibatch_attention(
        self,
        query: torch.Tensor,
        key: torch.Tensor,
        value: torch.Tensor,
        key_multiplicities: Optional[torch.Tensor]
    ):
        warnings.warn("Softmax flash attention does not support minibatch aggregation. Falling back to naive implementation.")
        return super().compute_minibatch_attention(query, key, value, key_multiplicities)


class ScaledDotProductSigmoidSetAttention(ScaledDotProductSetAttention):
    def compute_immediate_attention(
        self,
        query: torch.Tensor,
        key: torch.Tensor,
        value: torch.Tensor,
        key_multiplicities: Optional[torch.Tensor]
    ):
        scale = self.scale or 1.0/math.sqrt(query.shape[-1])
        attention = query @ key.transpose(-2, -1) * scale
        attention = torch.sigmoid(attention)
        if key_multiplicities is not None:
            attention = attention * key_multiplicities
        attention = torch.dropout(attention, self.dropout_p, self.training)
        return attention @ value

    def compute_minibatch_attention(
        self,
        query: torch.Tensor,
        key: torch.Tensor,
        value: torch.Tensor,
        key_multiplicities: Optional[torch.Tensor]
    ):
        scale = self.scale or 1.0/math.sqrt(query.shape[-1])
        attention = query @ key.transpose(-2, -1) * scale
        attention = torch.sigmoid(attention)
        if key_multiplicities is not None:
            attention = attention * key_multiplicities.float().unsqueeze(-2)
        attention = torch.dropout(attention, self.dropout_p, self.training)
        return {
            "attention": attention @ value
        }

    def initial_state(self):
        return {"attention": 0.0}

    def get(self, state: Any):
        return state["attention"]

scaled_dot_product_attentions = {
    "softmax": ScaledDotProductSoftmaxSetAttention,
    "softmax-flash": ScaledDotProductSoftmaxFlashSetAttention,
    "sigmoid": ScaledDotProductSigmoidSetAttention
}
