import abc
import copy
import itertools
import lightning as L
import numpy as np
from transformers import PretrainedConfig, PreTrainedModel
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.utils.checkpoint
from typing import Any, Callable, Generic, List, Literal, Iterable, Optional, Tuple, TypeVar, Union

from . import CrossSetAttentionBlock, SelfSetAttentionBlock, Slots
from .scaled_dot_product_attention import ScaledDotProductSetAttention
from ..api.attention import AttentionInterceptor
from ..api.checkpointing import MinibatchCheckpointableMixin
from ..api.modules import SetPoolingLayer, SetTransformerEncoderLayer
from ..api.types import Minibatch, Minibatches, NoState
from ..attention_attribution import AttentionAttributionContext
from ..checkpointing import MinibatchContext

class SetTransformerEncoder(nn.Module):
    """
    The Set Transformer encoder.
    """
    def __init__(
        self,
        encoder_layer: SetTransformerEncoderLayer,
        num_layers: int,
        norm: Optional[nn.Module] = None,
        use_activation_checkpointing: bool = False,
        use_reentrant_checkpointing: bool = False
    ):
        super().__init__()
        self.layers = nn.ModuleList([copy.deepcopy(encoder_layer) for _ in range(num_layers)])
        self.norm = norm
        self.use_activation_checkpointing = use_activation_checkpointing
        self.use_reentrant_checkpointing = use_reentrant_checkpointing

    def forward(
        self,
        src: Union[torch.Tensor, Iterable[torch.Tensor]],
        multiplicities: Optional[torch.Tensor] = None,
    ):
        output = src
        for layer in self.layers:
            if self.use_activation_checkpointing:
                output = torch.utils.checkpoint.checkpoint(
                    layer,
                    output,
                    multiplicities,
                    use_reentrant=self.use_reentrant_checkpointing
                )
            else:
                output = layer(output, multiplicities)
        if self.norm is not None:
            output = self.norm(output)
        return output

PreprocessorType = TypeVar("PreprocessorType", bound=Callable[[torch.Tensor, Optional[torch.Tensor], int], Tuple[torch.Tensor, Optional[torch.Tensor]]])
PrepoolerType = TypeVar("PrepoolerType", bound=nn.Module)
EncoderType = TypeVar("EncoderType", bound=SetTransformerEncoder)
PoolerType = TypeVar("PoolerType", bound=SetPoolingLayer)
DecoderType = TypeVar("DecoderType", bound=nn.Module)

class SetTransformerConfig(PretrainedConfig):
    """
    Configuration class for the Set Transformer.
    """
    model_type = "set_transformer"

    def __init__(
        self,
        gradient_minibatch_limit: Optional[int] = None,
        gradient_estimation_method: Literal["biased", "unbiased"] = "unbiased",
        persist_graph: bool = True,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.gradient_minibatch_limit = gradient_minibatch_limit
        self.gradient_estimation_method = gradient_estimation_method
        self.persist_graph = persist_graph

class SetTransformerModel(
    PreTrainedModel,
    L.LightningModule,
    MinibatchCheckpointableMixin,
    Generic[
        PreprocessorType,
        PrepoolerType,
        EncoderType,
        PoolerType,
        DecoderType
    ]
):
    """
    A Set Transformer model comprised of an encoder and an optional decoder. This
    model guarantees correct unbiased gradient estimation computation.
    """
    config_class = SetTransformerConfig

    def __init__(
        self,
        config: SetTransformerConfig,
        preprocessor: Optional[PreprocessorType] = None,
        prepooler: Optional[PrepoolerType] = None,
        encoder: Optional[EncoderType] = None,
        pooler: Optional[PoolerType] = None,
        decoder: Optional[DecoderType] = None
    ):
        super().__init__(config)
        self.preprocessor: Optional[PreprocessorType] = preprocessor
        self.prepooler: Optional[PrepoolerType] = prepooler
        self.encoder: Optional[EncoderType] = encoder
        self.pooler: Optional[PoolerType] = pooler
        self.decoder: Optional[DecoderType] = decoder

        self.automatic_optimization = False
        self._minibatch_context = MinibatchContext(
            gradient_minibatch_limit=self.config.gradient_minibatch_limit,
            gradient_estimation_method=self.config.gradient_estimation_method,
            persist_graph=self.config.persist_graph
        )

        self.attention_module_map = None
        self.attention_module_reverse_map = None

    def forward(
        self,
        src: Union[torch.Tensor, Minibatches],
        src_multiplicities: Optional[torch.Tensor] = None,
        minibatch_size: Optional[int] = None,
        *,
        state: Optional[Any] = NoState
    ):
        """
        A generalized forward pass for the Set Transformer. This method ensures that minibatches
        are forwarded in a generative process an that gradient checkpoints are used appropriately.
        """
        if state is not NoState and self.prepooler is None:
            raise ValueError("Cannot use stateful mode without a prepooler.")

        if minibatch_size == 0:
            minibatch_size = None

        # Convert to minibatch mode if minibatch size is provided
        if isinstance(src, torch.Tensor) and minibatch_size is not None:
            src_full = src
            src = ({
                "src": src_full[..., i:i+minibatch_size, :],
                "src_multiplicities": (src_multiplicities[..., i:i+minibatch_size] if src_multiplicities is not None else None)
            } for i in range(0, src_full.shape[-2], minibatch_size))

        # If we're in a stateful context and the input is a tensor, we need to treat it as a minibatch
        elif isinstance(src, torch.Tensor) and state is not NoState:
            src_full = src # Store a copy because we overwrite src with a generator
            src = ({
                "src": src_full,
                "src_multiplicities": src_multiplicities
            },)

        # Step 1.
        # If we have a preprocessor, pipe in the set minibatches.
        if self.preprocessor is not None:
            if isinstance(src, torch.Tensor):
                src, src_multiplicities = self.preprocessor(src, src_multiplicities, 0)
            else:
                src = (
                    dict(zip(("src", "src_multiplicities"), self.preprocessor(mb["src"], mb.get("src_multiplicities", None), i))) # type: ignore
                    for i, mb in enumerate(src)
                )

        # Step 2.
        # If we have a prepooler, pool the set everything together.
        if self.prepooler is not None:
            # If we're in a stateful context and the input is a tensor, we need to treat it as a minibatch
            if state is not NoState:
                src, state = self.prepooler(src, src_multiplicities, state=state)
            else:
                src = self.prepooler(src, src_multiplicities)

        # Step 3.
        # If we have an encoder, encode the set.
        if self.encoder is not None:
            if self.prepooler is None:
                # If we don't have a prepooler, we need to pass the multiplicities
                src = self.encoder(src, src_multiplicities)
            else:
                src = self.encoder(src)

        # Step 4.
        # If we have a pooler, pool the current set.
        if self.pooler is not None:
            if self.prepooler is None:
                # If we don't have a prepooler, we need to pass the multiplicities
                src = self.pooler(src, src_multiplicities) # type: ignore
            else:
                src = self.pooler(src)

        # Step 5.
        # If we have a decoder, decode the current set.
        if self.decoder is not None:
            if isinstance(src, torch.Tensor):
                src = self.decoder(src)
            else:
                src = ({"src": self.decoder(mb["src"])} for mb in src)

        # Final checkpoint
        if isinstance(src, torch.Tensor):
            src = self.minibatch_checkpoint_pipe(src, "post")
        else:
            src = self.minibatch_checkpoint_pipe(src, "post", extract_reference_tensor=lambda x: x["src"])

        if not isinstance(src, torch.Tensor):
            src = torch.cat([mb["src"] for mb in src], -2) # type: ignore

        assert isinstance(src, torch.Tensor)
        return self.forward_post(src)

    def forward_post(self, src: torch.Tensor):
        """
        Post processing for the forward pass.
        """
        return src

    def generic_step(self, mode: str, batch) -> torch.Tensor:
        raise NotImplementedError()

    def training_step(self, batch):
        # This is to allow overriding later
        setattr(self, "last_batch", batch)
        with self._minibatch_context as ctx:
            # Perform manual backward so that we can correct the gradients
            optimizer = self.optimizers()
            optimizer.zero_grad() # type: ignore
            loss = self.generic_step("train", batch)
            self.manual_backward(loss)
            for n, p in self.named_parameters():
                if p.grad is not None and torch.any(torch.isnan(p.grad)):
                    torch.save({
                        "state_dict": self.state_dict(),
                        "input": getattr(self, "last_batch"),
                        "loss": loss
                    }, "crash_state.pt")
                    raise ValueError(f"Gradient for {n} is NaN. Loss was: {loss}")
            ctx.correct_gradients()
            optimizer.step() # type: ignore
            return loss

    def validation_step(self, batch):
        return self.generic_step("val", batch)

    def test_step(self, batch):
        return self.generic_step("test", batch)

    def attribution_step(self, batch):
        return self.generic_step("attribution", batch)

    def attention_attribution(self, batch, show_progress: bool = False, return_raw_scores: bool = False):
        if self.attention_module_map is None:
            self._map_attention_modules()
        assert self.attention_module_map is not None and self.attention_module_reverse_map is not None
        with AttentionAttributionContext([a for m in self.attention_module_map.values() for a in m]) as ctx:
            minibatch_size = self.config.minibatch_size
            try:
                # Minibatching is currently not supported.
                self.config.minibatch_size = 0
                scores = ctx.compute(self.attribution_step, batch, show_progress=show_progress)
                scores_by_module = {}
                score_iter = iter(scores)
                for module, submodules in self.attention_module_map.items():
                    scores_by_module[module] = {}
                    for submodule, s in zip(submodules, score_iter):
                        scores_by_module[module][submodule] = s
            finally:
                # Restore minibatch_size
                self.config.minibatch_size = minibatch_size

        if return_raw_scores:
            return scores_by_module
        else:
            return self._aggregate_attention_scores(scores_by_module)

    def _aggregate_attention_scores(self, scores_by_module):
        attention = []
        for module, submodules in scores_by_module.items():
            scores = list(submodules.values())
            if len(scores) == 1:
                scores = scores[0]
            elif len(scores) == 2:
                scores = scores[0]
                # scores = (scores[0] + scores[1].transpose(-1, -2)) / 2.0
            else:
                raise ValueError(f"Expected 1 or 2 scores, got {len(scores)}")
            scores = scores.detach().cpu().numpy()
            scores = np.clip(scores, 0.0, None)
            scores = np.sum(scores, (1, 2)) # sum across heads and rows
            attention.append(scores)
        attention = np.stack(attention, 0)
        attention = np.sum(attention, 0) # sum across layers
        attention = attention / np.max(attention) # normalize by max score
        return attention

    def _map_attention_modules(self):
        self.attention_module_map = {}
        self.attention_module_reverse_map = {}
        if self.prepooler is not None:
            self.attention_module_map[self.prepooler] = self._find_attention_modules(self.prepooler)
        elif self.encoder is not None:
            for encoder_layer in self.encoder.layers:
                self.attention_module_map[encoder_layer] = self._find_attention_modules(encoder_layer)
        if self.prepooler is None and self.pooler is not None:
            self.attention_module_map[self.pooler] = self._find_attention_modules(self.pooler)
        to_remove = []
        for module, attention_modules in self.attention_module_map.items():
            if len(attention_modules) == 0:
                to_remove.append(module)
        for module in to_remove:
            del self.attention_module_map[module]
        for module, attention_modules in self.attention_module_map.items():
            for attention_module in attention_modules:
                self.attention_module_reverse_map[attention_module] = module

    def _find_attention_modules(self, module: nn.Module):
        attention_modules = []
        for submodule in module.modules():
            if isinstance(submodule, AttentionInterceptor):
                attention_modules.append(submodule)
        return attention_modules

# Set Transformer Encoder Layers -------------------------------------------------------------------

class SetAttentionBlock(SetTransformerEncoderLayer, SelfSetAttentionBlock):
    def __init__(
        self,
        d_model: int,
        nhead: int,
        dim_feedforward: Optional[int] = None,
        dropout: float = 0.1,
        activation: Union[str, Callable[[torch.Tensor], torch.Tensor]] = F.relu,
        attention_method: Optional[ScaledDotProductSetAttention] = None,
        layer_norm: bool = True,
        layer_norm_eps: float = 1e-5,
        norm_first: bool = True,
        bias: bool = True,
        d_head: Optional[int] = None,
        device: Optional[torch.device] = None,
        dtype: Optional[torch.dtype] = None
    ):
        super().__init__(
            d_model=d_model,
            nhead=nhead,
            dim_feedforward=dim_feedforward,
            dropout=dropout,
            activation=activation,
            attention_method=attention_method,
            layer_norm=layer_norm,
            layer_norm_eps=layer_norm_eps,
            norm_first=norm_first,
            bias=bias,
            d_head=d_head,
            device=device,
            dtype=dtype
        )

    def forward_immediate(
        self,
        src: torch.Tensor,
        src_multiplicities: Optional[torch.Tensor] = None,
    ):
        return SelfSetAttentionBlock.forward(self, src, src_multiplicities)


class InducedSetAttentionBlock(SetTransformerEncoderLayer):
    def __init__(
        self,
        d_model: int,
        nhead: int,
        num_slots: int,
        slot_type: Literal["random", "deterministic"] = "deterministic",
        dim_feedforward: Optional[int] = None,
        dropout: float = 0.1,
        activation: Union[str, Callable[[torch.Tensor], torch.Tensor]] = F.relu,
        attention_method: Optional[ScaledDotProductSetAttention] = None,
        slot_attention_method: Optional[ScaledDotProductSetAttention] = None,
        layer_norm: bool = True,
        layer_norm_eps: float = 1e-5,
        norm_first: bool = True,
        bias: bool = True,
        d_head: Optional[int] = None,
        device: Optional[torch.device] = None,
        dtype: Optional[torch.dtype] = None
    ):
        super().__init__()
        self.mab1 = CrossSetAttentionBlock(
            d_model=d_model,
            nhead=nhead,
            dim_feedforward=dim_feedforward,
            dropout=dropout,
            activation=activation,
            attention_method=slot_attention_method,
            layer_norm=layer_norm,
            layer_norm_eps=layer_norm_eps,
            norm_first=norm_first,
            bias=bias,
            d_head=d_head,
            device=device,
            dtype=dtype
        )
        self.mab2 = CrossSetAttentionBlock(
            d_model=d_model,
            nhead=nhead,
            dim_feedforward=dim_feedforward,
            dropout=dropout,
            activation=activation,
            attention_method=attention_method,
            layer_norm=layer_norm,
            layer_norm_eps=layer_norm_eps,
            norm_first=norm_first,
            bias=bias,
            d_head=d_head,
            device=device,
            dtype=dtype
        )
        self.slots = Slots(d_model, num_slots, slot_type=slot_type)

    def forward_immediate(
        self,
        src: torch.Tensor,
        src_multiplicities: Optional[torch.Tensor] = None,
    ):
        batch_size = src.shape[0]
        return self.mab2(
            src,
            self.mab1(
                self.slots(batch_size),
                src,
                src_multiplicities
            )
        )

    def forward_minibatch(
        self,
        src: Minibatches
    ) -> Minibatches:
        # We need to persist the multiplicities
        persisted_src = []

        # Peek at first minibatch to get batch size
        src = iter(src)
        first_minibatch = next(src)
        batch_size = first_minibatch["src"].shape[0] # type: ignore

        # Reinsert the first minibatch
        src = itertools.chain((first_minibatch,), src)
        first_minibatch = None # free

        # Compute induced attention while presisting src
        induced_attention = self.mab1(
            self.slots(batch_size),
            (self._minibatch_map_and_persist(mb, persisted_src) for mb in src),
        )

        return ({
            "src": self.mab2(
                mb["y"],
                induced_attention,
            ),
            "src_multiplicities": mb["y_multiplicities"]
        } for mb in persisted_src)

    def _minibatch_map_and_persist(self, src: Minibatch, persisted_src: List[Minibatch]):
        y = src.pop("src")
        y_multiplicities = src.pop("src_multiplicities", None)
        src = {
            "y": y,
            "y_multiplicities": y_multiplicities
        }
        persisted_src.append(src)
        return src


class PoolingByMultiheadSetAttention(SetPoolingLayer):

    def __init__(
        self,
        d_model: int,
        nhead: int,
        num_slots: int,
        slot_type: Literal["random", "deterministic"] = "deterministic",
        dim_feedforward: Optional[int] = None,
        dropout: float = 0.1,
        activation: Union[str, Callable[[torch.Tensor], torch.Tensor]] = F.relu,
        attention_method: Optional[ScaledDotProductSetAttention] = None,
        layer_norm: bool = True,
        layer_norm_eps: float = 1e-5,
        norm_first: bool = True,
        bias: bool = True,
        d_head: Optional[int] = None,
        device: Optional[torch.device] = None,
        dtype: Optional[torch.dtype] = None
    ):
        super().__init__()
        self.mab = CrossSetAttentionBlock(
            d_model=d_model,
            nhead=nhead,
            dim_feedforward=dim_feedforward,
            dropout=dropout,
            activation=activation,
            attention_method=attention_method,
            layer_norm=layer_norm,
            layer_norm_eps=layer_norm_eps,
            norm_first=norm_first,
            bias=bias,
            d_head=d_head,
            device=device,
            dtype=dtype
        )
        self.slots = Slots(d_model, num_slots, slot_type=slot_type)

        if isinstance(activation, str):
            self.activation = getattr(F, activation)
        else:
            self.activation = activation

        dim_feedforward = dim_feedforward or 4*d_model
        self.linear1 = nn.Linear(d_model, dim_feedforward, bias=bias, device=device, dtype=dtype)
        self.linear2 = nn.Linear(dim_feedforward, d_model, bias=bias, device=device, dtype=dtype)
        self.dropout = nn.Dropout(dropout)

    def _feedforward(self, src: torch.Tensor) -> torch.Tensor:
        return self.linear2(self.dropout(self.activation(self.linear1(src))))

    def forward_immediate(
        self,
        src: torch.Tensor,
        src_multiplicities: Optional[torch.Tensor] = None,
    ):
        batch_size = src.shape[0]
        return self.mab(
            self.slots(batch_size),
            self._feedforward(src),
            src_multiplicities
        )

    def forward_minibatch(
        self,
        src: Minibatches,
        state: Optional[Any]
    ) -> Tuple[torch.Tensor, Any]:
        # Peek at first minibatch to get batch size
        src = iter(src)
        first_minibatch = next(src)
        batch_size = first_minibatch["src"].shape[0] # type: ignore

        # Reinsert the first minibatch
        src = itertools.chain((first_minibatch,), src)
        first_minibatch = None # free

        # Compute attention
        return self.mab(
            self.slots(batch_size),
            ({
                "y": self._feedforward(mb["src"]), # type: ignore
                "y_multiplicities": mb.get("src_multiplicities", None)
            } for mb in src),
            state=state
        )
